#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

#define BLOCK_SIZE 256
#define UNROLL_FACTOR 4

struct PixelInfo {
    int batch_idx;
    int d, h, w;
};

__device__ PixelInfo compute_pixel_info(int idx, int depth, int height, int width) {
    PixelInfo pi;
    pi.batch_idx = idx / (depth * height * width);
    int rem = idx % (depth * height * width);
    pi.d = rem / (height * width);
    rem %= height * width;
    pi.h = rem / width;
    pi.w = rem % width;
    return pi;
}

template <typename scalar_t, int CHANNELS>
__global__ void fused_softmax_sigmoid_kernel(
    const scalar_t* __restrict__ input,
    scalar_t* __restrict__ output,
    int batch, int depth, int height, int width) {
    
    const int spatial = depth * height * width;
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= batch * spatial) return;

    const PixelInfo pi = compute_pixel_info(idx, depth, height, width);
    const int base = (pi.batch_idx * CHANNELS * spatial) + (pi.d * height * width + pi.h * width + pi.w);
    const int stride = spatial;

    scalar_t max_val = -INFINITY;
    scalar_t sum_exp = 0.0f;

    #pragma unroll
    for (int c = 0; c < CHANNELS; c += UNROLL_FACTOR) {
        scalar_t vals[UNROLL_FACTOR];
        #pragma unroll
        for (int u = 0; u < UNROLL_FACTOR; ++u) {
            vals[u] = input[base + (c + u) * stride];
            max_val = max(max_val, vals[u]);
        }
        #pragma unroll
        for (int u = 0; u < UNROLL_FACTOR; ++u) {
            sum_exp += exp(vals[u] - max_val);
        }
    }

    #pragma unroll
    for (int c = 0; c < CHANNELS; c += UNROLL_FACTOR) {
        #pragma unroll
        for (int u = 0; u < UNROLL_FACTOR; ++u) {
            const int pos = base + (c + u) * stride;
            scalar_t softmax = exp(input[pos] - max_val) / sum_exp;
            output[pos] = 1.0f / (1.0f + exp(-softmax));
        }
    }
}

template <typename scalar_t>
__global__ void dynamic_fused_softmax_sigmoid_kernel(
    const scalar_t* __restrict__ input,
    scalar_t* __restrict__ output,
    int channels, int batch, int depth, int height, int width) {
    
    const int spatial = depth * height * width;
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= batch * spatial) return;

    const PixelInfo pi = compute_pixel_info(idx, depth, height, width);
    const int base = (pi.batch_idx * channels * spatial) + (pi.d * height * width + pi.h * width + pi.w);
    const int stride = spatial;

    scalar_t max_val = -INFINITY;
    for (int c = 0; c < channels; ++c) {
        max_val = max(max_val, input[base + c * stride]);
    }

    scalar_t sum_exp = 0.0f;
    #pragma unroll 4
    for (int c = 0; c < channels; ++c) {
        sum_exp += exp(input[base + c * stride] - max_val);
    }

    #pragma unroll 4
    for (int c = 0; c < channels; ++c) {
        const int pos = base + c * stride;
        scalar_t softmax = exp(input[pos] - max_val) / sum_exp;
        output[pos] = 1.0f / (1.0f + exp(-softmax));
    }
}

torch::Tensor forward(torch::Tensor input, int stride, int padding, int output_padding, bool bias_flag, torch::Tensor conv_transpose, torch::Tensor conv_transpose_bias) {
    auto x = torch::conv_transpose3d(input, conv_transpose, bias_flag ? conv_transpose_bias : torch::Tensor(), stride, padding, output_padding);
    
    const int batch = x.size(0);
    const int channels = x.size(1);
    const int depth = x.size(2);
    const int height = x.size(3);
    const int width = x.size(4);

    auto output = torch::empty_like(x);
    const int total_pixels = batch * depth * height * width;
    const int blocks = (total_pixels + BLOCK_SIZE - 1) / BLOCK_SIZE;

    AT_DISPATCH_FLOATING_TYPES(x.scalar_type(), "fused_forward", ([&] {
        if (channels == 32) {
            fused_softmax_sigmoid_kernel<scalar_t, 32><<<blocks, BLOCK_SIZE>>>(
                x.data_ptr<scalar_t>(), output.data_ptr<scalar_t>(), batch, depth, height, width);
        } else if (channels == 64) {
            fused_softmax_sigmoid_kernel<scalar_t, 64><<<blocks, BLOCK_SIZE>>>(
                x.data_ptr<scalar_t>(), output.data_ptr<scalar_t>(), batch, depth, height, width);
        } else {
            dynamic_fused_softmax_sigmoid_kernel<scalar_t><<<blocks, BLOCK_SIZE>>>(
                x.data_ptr<scalar_t>(), output.data_ptr<scalar_t>(), channels, batch, depth, height, width);
        }
    }));

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused ConvTranspose3D with Optimized Softmax-Sigmoid");
}